<html>
  <head>
    <!-- Global site tag (gtag.js) - Google Analytics -->
    <script async src="https://www.googletagmanager.com/gtag/js?id=default-key"></script>
    <script>
      const google = {"key":"default-key","sampleRate":100};
      window.dataLayer = window.dataLayer || [];
      function gtag() { dataLayer.push(arguments); }
      gtag('js', new Date());
      gtag('config', google.key);
      gtag('create', google.key, { 'siteSpeedSampleRate': google.sampleRate });
    </script>
    <meta charset="utf-8">
    <link rel="shortcut icon" href="/static/images/favicon.png">
    <link href="/static/dist/main.css" type="text/css" rel="stylesheet"/>
    <title>Amundsen - Data Discovery Portal</title>
  </head>
  <body>
    <div id="content"/>
    <script src="/static/dist/vendors.js" type="text/javascript"></script>
    <script src="/static/dist/main.js" type="text/javascript"></script>
  </body>
  <script type="text/javascript">
    // Feature detects Navigation Timing API support.
    if (window.performance) {
      // Gets the number of milliseconds since page load
      // (and rounds the result since the value must be an integer).
      const timeSincePageLoad = Math.round(performance.now());

      // Sends the timing event to Google Analytics.
      gtag('event', 'timing_complete', {
        'name': 'load',
        'value': timeSincePageLoad,
        'event_category': 'JS Dependencies'
      });
    }
  </script>
</html>
